/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
ldse@hendriklipka.de
*/

package de.hendriklipka.ldse.server;

import java.util.Vector;
import java.io.File;

import com.sun.xml.parser.Resolver;
import com.sun.xml.tree.XmlDocument;
import org.xml.sax.InputSource;
import org.w3c.dom.*;

import HTTPClient.*;

import de.hendriklipka.ldse.common.Log;
import de.hendriklipka.ldse.common.Spider;
import de.hendriklipka.ldse.server.data.URLInfo;
import de.hendriklipka.ldse.server.dataserver.DataClient;
import de.hendriklipka.ldse.server.indexer.IndexServer;
import de.hendriklipka.ldse.server.indexer.DocumentFactory;

public class Gatherer
{
  private class SpiderThread extends Thread
  {
    IndexServer _is=null;
    Spider _spider=null;
    public int _id=-1;
    
    public SpiderThread(IndexServer is, Spider spider, int id)
    {
      super("Server.SpiderThread");
      _is=is;
      _id=id;
      _spider=spider;
      _spider.start();
      setPriority(getPriority()+2);
    }
    
    public void run()
    {
      while ( _spider.hasMoreElements() )
      {
        try
        {
          Log.log("server","threads",2,"spider is looking for next URL");
          URLInfo ui=_spider.getNextElement();
          if (null==ui)
            break;
          // keine URL -> Fehler...
          if (null==ui._url)
          {
            Log.log("server","threads",0,"got no url, breaking out of loop");
            break;
          }
          // kein Typ.... nehmen wir mal "text/plain" an
          if (null==ui._type)
          {
            ui._type="text/plain";
          }
          Log.log("server","info",0,"putting URL: "+ui._urlStr);
          _is.putURL(ui);
        }
        catch (Exception ex) 
        {
          ex.printStackTrace(Log.getWriter("dataclient","error",0));
        }
      }
      Log.log("server","threads",2,"no more URLs from this spider");
    }
  }
  
  protected Vector _spiders=new Vector(5,5);
  
  public Gatherer(String cfgFile)
  {
    if (null==cfgFile)
        cfgFile="documentserver.xml";
    try
    {
      // Konfiguration einlesen (bis zur Config des Logwriters...)
	    InputSource	input = Resolver.createInputSource (new File (cfgFile));
	    XmlDocument	doc = XmlDocument.createXmlDocument (input, false);
      Element el=doc.getDocumentElement();
      NodeList cnl=el.getElementsByTagName("debug");
      Node configNode=cnl.item(0);
      Log.init(configNode);
      
      // DataClient erzeugen
      NodeList nl=el.getElementsByTagName("dataclient");
      Node node=nl.item(0);
      DataClient dc=new DataClient(node);
      
      // Indexer erzeugen
      nl=el.getElementsByTagName("documentfactory");
      node=nl.item(0);
      DocumentFactory.init(node);

      // Indexer erzeugen
      nl=el.getElementsByTagName("indexer");
      node=nl.item(0);
      IndexServer is=new IndexServer(node,dc);
      
      // Spider erzeugen und starten
      nl=el.getElementsByTagName("gatherer");
      node=nl.item(0);
      nl=node.getChildNodes();
      for (int i=0;i<nl.getLength();i++)
      {
        node=nl.item(i);
        if ((node.getNodeType()==node.ELEMENT_NODE)&&(node.getNodeName().equals("spider")))
        {
          NamedNodeMap nnl=node.getAttributes();
          Node attr=nnl.getNamedItem("class");
          String className=attr.getNodeValue();
          attr=nnl.getNamedItem("id");
          String idStr=attr.getNodeValue();
          int id=Integer.parseInt(idStr);
          try
          {
            Class cls=Class.forName(className);
            Object obj=cls.newInstance();
            Spider spider=(Spider)obj;
            spider.init(node,id); 
            dc.markSpider(id);
            // Spider starten (eigener Thread)
            SpiderThread st=new SpiderThread(is,spider,id);
            _spiders.addElement(st);
            st.start();
          }
          catch (ClassNotFoundException ex)
          {
            ex.printStackTrace();
          }
        }
        else if ((node.getNodeType()==node.ELEMENT_NODE)&&(node.getNodeName().equals("proxy")))
        {
          NamedNodeMap nnl=node.getAttributes();
          Node attr=nnl.getNamedItem("useproxy");
          String s=attr.getNodeValue();
          if (s.equals("no"))
            continue;
          attr=nnl.getNamedItem("proxyhost");
          String ph=attr.getNodeValue();
          attr=nnl.getNamedItem("proxyport");
          s=attr.getNodeValue();
          int pp=Integer.parseInt(s);
          HTTPConnection.setProxyServer(ph,pp);
          attr=nnl.getNamedItem("dontproxy");
          s=attr.getNodeValue();
          attr=nnl.getNamedItem("protocol");
          s=attr.getNodeValue();
        }
      }
      
      // alle Spider joinen
      Log.log("server","threads",0,"joining spiders");
      while (0!=_spiders.size())
      {
        SpiderThread st=(SpiderThread)_spiders.firstElement();
        st.join();
        _spiders.removeElementAt(0);
        int id=st._id;
        dc.sweepSpider(id);
      }
      Log.log("server","threads",0,"all spiders are dead");
      
      // DataClient schliessen
      dc.close();
    }
    catch (Throwable ex)
    {
      ex.printStackTrace(System.err);
      ex.printStackTrace(Log.getWriter("server","error",0));
    }
  }
  
  public static void main(String[] args)
  {
     String cfgFile=null;
     if (args.length>0)
        cfgFile=args[0];
   new Gatherer(cfgFile);
  }
}