/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms 
of the GNU General Public License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY 
WARRANTY; without even the implied warranty of MERCHANTABILITY or 
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License 
for more details. 

You should have received a copy of the GNU General Public License along with this 
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave, 
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
ldse@hendriklipka.de
*/

package de.hendriklipka.ldse.server.query;

import java.util.*;

import com.objectspace.jgl.Container;
import com.objectspace.jgl.SList;
import com.objectspace.jgl.OrderedSet;
import com.objectspace.jgl.BinaryPredicate;
import com.objectspace.jgl.HashMapIterator;
import com.objectspace.jgl.ForwardIterator;

import org.w3c.dom.*;

import de.hendriklipka.ldse.common.Log;
import de.hendriklipka.ldse.common.ResultDescriptor;
import de.hendriklipka.ldse.server.data.document.Document;
import de.hendriklipka.ldse.server.data.document.*;
import de.hendriklipka.ldse.server.common.*;
import de.hendriklipka.ldse.server.data.URLInfo;
import de.hendriklipka.ldse.server.indexer.DocumentFactory;
import de.hendriklipka.ldse.server.dataserver.DBConnection;
import de.hendriklipka.ldse.server.broker.BrokerQuery;

/**
  Behandelt erweiterte Anfragen
*/
public class ComplexQueryProcessor extends de.hendriklipka.ldse.server.query.QueryProcessor
{
  protected String[] _words=null;
  protected TripletList _triplets=null;
  protected TripletList[] _tripletWords=null;
  protected DocumentDescriptor[] _tempResult=null;
  protected WordData[] _wordData=null;
//  protected int _divisor=0;
  
	public ComplexQueryProcessor(Node query)
	{
		super(query);
	}

  // Woerter in Triplets zerlegen und als Liste speichern
  public void preprocessQuery()
  {
    _tripletWords=new TripletList[_words.length];
    TripletList list=new TripletList();
    for (int i=0;i<_words.length;i++)
    {
      TripletList tl=new TripletList(_words[i]);
      _tripletWords[i]=tl;
      list.addList(tl);
      QueryServer._queryWords++;
    }
    _triplets=list;
    Enumeration e=_triplets.getTriplets();
    while (e.hasMoreElements())
    {
      Triplet tr=(Triplet)e.nextElement();
//      _divisor+=tr.count*tr.count;
      QueryServer._queryTriplets++;
    }
  }

  // eigentliches Resultat erzeugen
  public void postprocessQuery(DBConnection con)
  {
    // 1. wurde schon in doQuery() abgehandelt...
    
    // 2. Relevanzen fuer die Dokumente berechnen 
    //   Relevanz des Wortes * Auftretensanzahl des Wortes im Dokument
    int count=0;
    Container result=null;

    if (null!=_limits)
    {
      if (!_limits._sorted&&(0==_limits._maxHits))
      {
        result=new SList();
      }
    }
    if (null==result)
    {
      result=new OrderedSet(ResultDescriptor.getPredicate(),true);
    }
    Log.log("queryserver","query",3,"result has "+_tempResult.length+" elements");
    for (int i=0;i<_tempResult.length;i++)
    {
      DocumentDescriptor dd=_tempResult[i];
      ResultDescriptor rd=dd.createResultDescriptor();
      double level=0;
      
      for (int j=0;j<dd._words.size();j++)
      {
        WordDocument wd=(WordDocument)dd._words.elementAt(j);
        level+=wd._count*wd._word._relevance;
      }
      
      // nur relevante Dokumente als Ergebnis liefern
      if (0==level)
        continue;
      if (null!=_limits)
      {
        if (_limits._minLevel<=level)
        {
          rd.setRelevance(level);
          result.add(rd);
          count++;
        }
      }
      else // kein Limit -> auf jeden Fall einfuegen
      {
        count++;
        rd.setRelevance(level);
        result.add(rd);
      }
    }
    if (null!=_limits)
    {
      if (!_limits._sorted&&(0!=_limits._maxHits))
      {
        _result=result;
      }
    }
    if (null==_result)
    {
      _result=new SList();
      ForwardIterator it=result.start();
      int i=0;
      while (!it.atEnd()&&(i<count))
      {
        ResultDescriptor rd=(ResultDescriptor)it.get();
        _result.add(rd);
        i++;
        it.nextElement();
      }
    }
    Log.log("queryserver","query",3,"final result has "+_result.size()+" elements");
  }

  public void init()
  {
    String words=new String();;
    NodeList subs=_query.getChildNodes();
    for (int j=0;j<subs.getLength();j++)
    {
      Node attr=subs.item(j);
      if (attr.getNodeType()!=attr.TEXT_NODE)
        continue;
      Text val=(Text)attr;
      String s=val.getData().trim();
      if (0!=s.length())
      {
        words+=" "+s;
      }
    }
    words=words.trim();
    Log.log("queryserver","query",2,"final words: "+words);
    Vector v=new Vector();
    StringTokenizer st=new StringTokenizer(words);
    while (st.hasMoreElements())
    {
      String s=(String)st.nextElement();
      Log.log("queryserver","query",2,"adding word: "+s);
      v.addElement(s);
    }
    _words=new String[v.size()];
    for (int i=0;i<v.size();i++)
    {
      _words[i]=(String)v.elementAt(i);
    }
  }

  public void doQuery(DBConnection con)
  {
    WordData[] wda=con.getWordsForTriplets(_triplets);
    _wordData=wda;
    
    // 1. Relevanzen fuer die einzelnen Woerter berechnen
    //   Vergleich der Woerter
    for (int i=0;i<_wordData.length;i++)
    {
        String word=_wordData[i]._word;
        TripletList tr=new TripletList(word);
        double level=0;
        // wir addieren nur den best match von allen Woertern der Anfrage
        double max=0;
        for (int t=0;t<_tripletWords.length;t++)
        {
          TripletList list=_tripletWords[t];
          double v=correlateWords(list,tr);
          if (v>max)
            max=v;
        }
        _wordData[i]._relevance=max;
    }
    
    Hashtable docs=new Hashtable();
    for (int i=0;i<wda.length;i++)
    {
        WordData wd=wda[i];
        if (0==wd._relevance)
          continue;
        con.fillWordData(wd);
        for (int j=0;j<wd._docs.length;j++)
        {
            DocumentDescriptor dd=wd._docs[j]._docDesc;
            Integer id=new Integer(dd.getID());
            Object o=docs.get(id);
            if (null!=o)
            {
                DocumentDescriptor dd2=(DocumentDescriptor)o;
                for (int k=0;k<dd._words.size();k++)
                {
                    dd2._words.addElement(dd._words.elementAt(k));
                }
            }
            else
            {
                docs.put(id,dd);
            }
        }
    }
    _tempResult=new DocumentDescriptor[docs.size()];
    int i=0;
    Enumeration e=docs.elements();
    while (e.hasMoreElements())
    {
        DocumentDescriptor dd=(DocumentDescriptor)e.nextElement();
        _tempResult[i]=dd;
        i++;
    }
  }
  
  public BrokerQuery createBrokerQuery()
  {
    String s="<simple>\n";
    for (int i=0;i<_words.length;i++)
    {
      s+=_words[i]+"\n";
    }
    s+="</simple>\n";

    return new BrokerQuery(s,this,_triplets,_limits);
  }
    
  public void close()
  {
  }
  
  // 1. Wort stammt aus der Frage, das 2. aus dem Dokument
  public double correlateWords(TripletList word1, TripletList word2)
  {
    double level=0;
    HashMapIterator it=word1.getIterator();
    int wl1=0;
    int count=0;
    while (!it.atEnd())
    {
      Triplet tr=(Triplet)it.value();
      wl1+=tr.count; // Wortlaenge des 1. Wortes
      // schauen, ob dieses Dokument es hat
      Triplet tr2=word2.getTriplet(tr);
      // ja -> Relevanz entsprechend erhoehen
      if (null!=tr2)
      {
        level+=(tr.count*tr2.count);
        count++;
      }
      it.advance();
    }
    // nur wenn im Dokument-Wort mind. 50% der Triplets aus dem Frage-Wort sind
    // wird der Wert genommen
    int wl2=word2.getTripletCount();
    level=(2*level)/(wl1+wl2);
    if (count>(word1.getTripletLength()/2))
    {
      return level;
    }
    return 0;
  }
    /*
  protected String[] _words=null;
  protected TripletList _triplets=null;
  protected TripletList[] _tripletWords=null;
  protected DocumentDescriptor[] _tempResult=null;
  protected Container result=null;
  protected int _resultCount=0;
  
  protected final static int MAXTHREADS=1;
  protected int _nextThread=0;
  protected QueryThread[] _threads=null;
  
	public ComplexQueryProcessor(Node query)
	{
		super(query);
	}

  // Woerter in Triplets zerlegen und als Liste speichern
  public void preprocessQuery()
  {
    _tripletWords=new TripletList[_words.length];
    TripletList list=new TripletList();
    for (int i=0;i<_words.length;i++)
    {
      QueryServer._queryWords++;
      TripletList tl=new TripletList(_words[i]);
      _tripletWords[i]=tl;
      list.addList(tl);
    }
    QueryServer._queryTriplets=list.getTripletCount();
    _triplets=list;
  }

  // eigentliches Resultat erzeugen
  public void postprocessQuery()
  {
    if (null!=_limits)
    {
      if (!_limits._sorted&&(0==_limits._maxHits))
      {
        result=new SList();
      }
      else
      {
        System.out.println("SimpleQueryProcessor: sorted or maxHits...");
        System.out.println("SimpleQueryProcessor: sorted: "+_limits._sorted);
        System.out.println("SimpleQueryProcessor: maxHits: "+_limits._maxHits);
      }
    }
    else
    {
      System.out.println("SimpleQueryProcessor: no limits...");
    }
    if (null==result)
    {
      System.out.println("SimpleQueryProcessor: OrderedSet");
      result=new OrderedSet(ResultDescriptor.getPredicate(),true);
    }

    Log.log("queryserver","query",3,"result has "+_tempResult.length+" elements");
    _threads=new QueryThread[_tempResult.length];
    
    for (int i=0;i<_tempResult.length;i++)
    {
      DocumentDescriptor dd=_tempResult[i];
      ResultDescriptor rd=dd.createResultDescriptor();

      URLInfo ui=dd.getURLInfo();
      
      // Threads starten
      QueryThread qt=new QueryThread(ui,rd);
      _threads[i]=qt;
    }
    // Threads starten ...
    int n=MAXTHREADS;
    if (n>_threads.length)
      n=_threads.length;
    _nextThread=n;
    for (int i=0;i<n;i++)
    {
      _threads[i].start();
    }
    
    // ... und warten, bis alle fertig sind...
    for (int i=0;i<_threads.length;i++)
    {
      try
      {
        QueryThread qt=_threads[i];
        qt.join();
      }
      catch (Exception ex) {}
    }
    _result=result;
    // Ergebniss in Array uebertragen
    Log.log("queryserver","query",3,"final result has "+_resultCount+" elements");
  }

  // Anfrage parsen
  public void init()
  {
    String words=new String();;
    NodeList subs=_query.getChildNodes();
    for (int j=0;j<subs.getLength();j++)
    {
      Node attr=subs.item(j);
      if (attr.getNodeType()!=attr.TEXT_NODE)
        continue;
      Text val=(Text)attr;
      String s=val.getData().trim();
      if (0!=s.length())
      {
        words+=" "+s;
      }
    }
    words=words.trim();
    Log.log("queryserver","query",2,"final words: "+words);
    Vector v=new Vector();
    StringTokenizer st=new StringTokenizer(words);
    while (st.hasMoreElements())
    {
      String s=(String)st.nextElement();
      Log.log("queryserver","query",2,"adding word: "+s);
      v.addElement(s);
    }
    _words=new String[v.size()];
    for (int i=0;i<v.size();i++)
    {
      _words[i]=(String)v.elementAt(i);
    }
  }

  public void doQuery(DBConnection con)
  {
    _tempResult=con.queryDocuments(_triplets,null);
  }
  
  public void close()
  {
  }

  // 1. Wort stammt aus der Frage, das 2. aus dem Dokument
  public double correlateWords(TripletList word1, TripletList word2)
  {
    double level=0;
    HashMapIterator it=word1.getIterator();
    int wl1=0;
    int count=0;
    while (!it.atEnd())
    {
      Triplet tr=(Triplet)it.value();
      wl1+=tr.count; // Wortlaenge des 1. Wortes
      // schauen, ob dieses Dokument es hat
      Triplet tr2=word2.getTriplet(tr);
      // ja -> Relevanz entsprechend erhoehen
      if (null!=tr2)
      {
        level+=(tr.count*tr2.count);
        count++;
      }
      it.advance();
    }
    // nur wenn im Dokument-Wort mind. 50% der Triplets aus dem Frage-Wort sind
    // wird der Wert genommen
    int wl2=word2.getTripletCount();
    level=(2*level)/(wl1+wl2);
    if (count>(word1.getTripletLength()/2))
    {
      return level;
    }
    return 0;
  }

  public BrokerQuery createBrokerQuery()
  {
    String s="<simple>\n";
    for (int i=0;i<_words.length;i++)
    {
      s+=_words[i]+"\n";
    }
    s+="</simple>\n";

    return new BrokerQuery(s,this,_triplets,_limits);
  }
  
  protected synchronized void threadFinished()
  {
    if (_nextThread<_threads.length)
    {
      Log.log("queryserver","query",3,"reading document #"+_nextThread);
      _threads[_nextThread].start();
      _nextThread++;
    }
  }
  
  class QueryThread extends Thread
  {
    protected double _result=0;
    protected URLInfo _urlInfo=null;
    protected ResultDescriptor _rd=null;
    
    public QueryThread(URLInfo urlInfo, ResultDescriptor rd)
    {
      super("QueryThread");
      _urlInfo=urlInfo;
      _rd=rd;
    }
    
    public void run()
    {
      Document doc=DocumentFactory.getDocument(_urlInfo);

      doc.init();
      
      Log.log("queryserver","query",3,"reading document for "+_urlInfo._urlStr);
      while (doc.hasMoreWords())
      {
        String word=doc.nextWord();
        TripletList tr=new TripletList(word);
        double level=0;
        // wir addieren nur den best match von allen Woertern der Anfrage
        double max=0;
        for (int t=0;t<_tripletWords.length;t++)
        {
          TripletList list=_tripletWords[t];
          double v=correlateWords(list,tr);
          if (v>max)
            max=v;
        }
        _result+=max;
      }
      
      // Ergebniss in Liste eintragen
      if (null!=_limits)
      {
        if ((_limits._minLevel<=_result)||(0==_limits._minLevel))
        {
          if (0<_result)
          {
            _rd.setRelevance(_result);
            result.add(_rd);
            _resultCount++;
          }
        }
      }
      else
      {
        _rd.setRelevance(_result);
        result.add(_rd);
        _resultCount++;
      }
      threadFinished();
    }
    
    public double getResult()
    {
      return _result;
    }
    
    public ResultDescriptor getResultDescriptor()
    {
      return _rd;
    }
  }
  */
}